import unittest
from unittest.mock import patch
from yt_fts.utils import split_subtitles

class TestSplitSubtitles(unittest.TestCase):
    def test_split_subtitles(self):
        # Setup
        video_id = "5XyayUs6J1M"
        # Call the function with the test parameters
        split_subtitles(video_id)

        # Add assertions here to check the behavior of the function
        # For example, you might check the output of the print statements

if __name__ == '__main__':
    unittest.main()


import click
import requests

from .config import get_config_path, get_db_path, get_or_make_chroma_path 
from .db_utils import *
from .download import *
from .list import list_channels
from .update import update_channel
from .utils import *
from rich.console import Console

YT_FTS_VERSION = "0.1.42"

@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(YT_FTS_VERSION, message='yt_fts version: %(version)s')
def cli():
    config_path = get_config_path()
    db_path = get_db_path()


# download
@cli.command( 
    help="""
    Download subtitles from a specified YouTube channel.

    You must provide the URL of the channel as an argument. The script will automatically extract the channel id from the URL.
    """
)
@click.argument("channel_url", required=True)
@click.option("-l", "--language", default="en", help="Language of the subtitles to download")
@click.option("-j", "--number-of-jobs", type=int, default=1, help="Optional number of jobs to parallelize the run")
def download(channel_url, language, number_of_jobs):
    console = Console()
    s = requests.session()


    # find out if the channel exists on the internet 
    with console.status("[bold green]Getting Channel ID...") as status:
        channel_url = validate_channel_url(channel_url)
        handle_reject_consent_cookie(channel_url, s)
        channel_id = get_channel_id(channel_url, s)
   
    if channel_id is None:
        console.print("[bold red]Error:[/bold red] Invalid channel URL or unable to extract channel ID.")
        return

    channel_exists = check_if_channel_exists(channel_id)

    if channel_exists:
        list_channels(channel_id)
        error = "[bold red]Error:[/bold red] Channel already exists in database."
        error += " Use update command to update the channel"
        console.print(error)
        return

    handle_reject_consent_cookie(channel_url, s)
    channel_name = get_channel_name(channel_id, s)
    
    if channel_name is None:
        console.print("[bold red]Error:[/bold red] The channel does not exist.")
        return

    foo = download_channel(channel_id, channel_name, language, number_of_jobs, s)
    if foo is None:
        console.print("[bold red]Error:[/bold red] Unable to download channel.")
        return
    else:
        console.print("[green]Download complete[/green]")


# list 
@cli.command( 
    help="""
    View library, transcripts and channel video list 
    """
)
@click.option("-t", "--transcript", default=None, help="Show transcript for a video")
@click.option("-c", "--channel", default=None, help="Show list of videos for a channel")
@click.option("-l", "--library", is_flag=True, help="Show list of channels in library")
def list(transcript, channel, library):

    from yt_fts.list import show_video_list, show_video_transcript

    if transcript:
        show_video_transcript(transcript)
        exit()
    elif channel:
        channel_id = get_channel_id_from_input(channel)
        show_video_list(channel_id)
    elif library:
        list_channels()
    else:
        list_channels()


# update
@cli.command( 
    help="""
    Updates a specified YouTube channel.

    You must provide the ID of the channel as an argument.
    Keep in mind some might not have subtitles enabled. This command
    will still attempt to download subtitles as subtitles are sometimes added later.
    """
)
@click.option("-c", "--channel", default=None, required=True, help="The name or id of the channel to update.")
@click.option("-l", "--language", default="en", help="Language of the subtitles to download")
@click.option("-j", "--number-of-jobs", type=int, default=1, help="Optional number of jobs to parallelize the run")
def update(channel, language, number_of_jobs):

    channel_id = get_channel_id_from_input(channel)
    channel_url = f"https://www.youtube.com/channel/{channel_id}/videos" 

    s = requests.session()
    handle_reject_consent_cookie(channel_url, s)

    channel_name = get_channel_name(channel_id, s)

    update_channel(channel_id, channel_name, language, number_of_jobs, s)


# Delete
@cli.command( 
    help="""
    Delete a channel and all its data. 

    You must provide the name or the id of the channel you want to delete as an argument. 

    The command will ask for confirmation before performing the deletion. 
    """
)
@click.option("-c", "--channel", default=None, required=True, help="The name or id of the channel to delete")
def delete(channel):

    channel_id = get_channel_id_from_input(channel)
    channel_name = get_channel_name_from_id(channel_id) 
    channel_url = f"https://www.youtube.com/channel/{channel_id}/videos"

    print(f"Deleting channel {channel_name}: {channel_url}")
    print("Are you sure you want to delete this channel and all its data?")
    confirm = input("y/n: ")

    if confirm == "y":
        delete_channel(channel_id)
        print(f"Deleted channel {channel_name}: {channel_url}")
    else:
        print("Exiting")


@cli.command(
        help="""
        export transcripts
        """
)
@click.option("-c", "--channel", default=None, required=True, help="The name or id of the channel to export transcripts for")
@click.option("-f", "--format", default="txt", help="The format to export transcripts to. Supported formats: txt, vtt")
def export(channel, format):

    from .export import export_channel_to_txt, export_channel_to_vtt
    console = Console()

    channel_id = get_channel_id_from_input(channel)

    if format == "txt":
        output_dir = export_channel_to_txt(channel_id)
    
    if format == "vtt":
        output_dir = export_channel_to_vtt(channel_id)

    if output_dir != None:
        console.print(f"Exported to [green][bold]{output_dir}[/bold][/green]")

# search
@cli.command(
        help="""
        Search for a specified text within a channel, a specific video, or across all channels.
        """
)
@click.argument("text", required=True)
@click.option("-c", "--channel", default=None, help="The name or id of the channel to search in.")
@click.option("-v", "--video", default=None, help="The id of the video to search in.")
@click.option("-l", "--limit", default=None, type=int, help="Number of results to return")
@click.option("-e", "--export", is_flag=True, help="Export search results to a CSV file.")
def search(text, channel, video, export, limit):

    from yt_fts.search import fts_search, print_fts_res  
    from yt_fts.export import export_fts 

    console = Console()

    if len(text) > 40:
        show_message("search_too_long")
        return

    if channel:
        scope = "channel"
    elif video:
        scope = "video"
    else:
        scope = "all"

    res = fts_search(text, scope, channel_id=channel, video_id=video, limit=limit)
    print_fts_res(res, text)

    if export:
        export_fts(text, scope, channel_id=channel, video_id=video)

    console.print(f"Query '{text}' ")
    console.print(f"Scope: {scope}")


# vsearch
@cli.command(
        help="""
            Vector search. Requires embeddings to be generated for the channel and environment variable OPENAI_API_KEY to be set.
        """
)
@click.argument("text", required=True)
@click.option("-c", "--channel", default=None, help="The name or id of the channel to search in")
@click.option("-v", "--video", default=None, help="The id of the video to search in.")
@click.option("-l", "--limit", default=10, help="Number of results to return")
@click.option("-e", "--export", is_flag=True, help="Export search results to a CSV file.")
@click.option("--openai-api-key", default=None, help="OpenAI API key. If not provided, the script will attempt to read it from the OPENAI_API_KEY environment variable.")
def vsearch(text, channel, video, limit, export, openai_api_key): 
    from openai import OpenAI
    from yt_fts.vector_search import search_chroma_db, print_vector_search_results
    from yt_fts.export import export_vector_search

    console = Console()

    if len(text) > 80:
        show_message("search_too_long")
        exit()

    # get api key for openai
    if  openai_api_key is None:
        openai_api_key = os.environ.get("OPENAI_API_KEY") 

    if openai_api_key is None:
        console.print("""
        [bold][red]Error:[/red][/bold] OPENAI_API_KEY environment variable not set, Run: 
                
                export OPENAI_API_KEY=<your_key> to set the key
                      """)
        return 

    openai_client = OpenAI(api_key=openai_api_key)


    if channel:
        scope = "channel"
    elif video:
        scope = "video"
    else:
        scope = "all"
    
    res = search_chroma_db(text, 
                           scope, 
                           channel_id=channel, 
                           video_id=video, 
                           limit=limit, 
                           openai_client=openai_client)
    
    print_vector_search_results(res, query=text)

    if export:    
        export_vector_search(res, text, scope)

    console.print(f"Query '{text}' ")
    console.print(f"Scope: {scope}")


# get-embeddings 
@cli.command( 
    help="""
    Generate embeddings for a channel using OpenAI's embeddings API.

    Requires an OpenAI API key to be set as an environment variable OPENAI_API_KEY.
    """
)
@click.option("-c", "--channel", default=None, help="The name or id of the channel to generate embeddings for")
@click.option("--openai-api-key", default=None, help="OpenAI API key. If not provided, the script will attempt to read it from the OPENAI_API_KEY environment variable.")
def get_embeddings(channel, openai_api_key):

    from yt_fts.db_utils import get_vid_ids_by_channel_id
    from yt_fts.embeddings import add_embeddings_to_chroma 
    from yt_fts.utils import split_subtitles, check_ss_enabled, enable_ss
    from openai import OpenAI

    console = Console()

    channel_id = get_channel_id_from_input(channel)

    # verify that embeddings have not already been created for the channel
    if check_ss_enabled(channel_id) == True:
        console.print("\n\t[bold][red]Error:[/red][/bold] Embeddings already created for this channel.\n")
        return


    # get api key for openai
    if  openai_api_key is None:
        openai_api_key = os.environ.get("OPENAI_API_KEY") 

    if openai_api_key is None:
        console.print("""
        [bold][red]Error:[/red][/bold] OPENAI_API_KEY environment variable not set, Run: 
                
                export OPENAI_API_KEY=<your_key> to set the key
                      """)
        return 

    openai_client = OpenAI(api_key=openai_api_key)

    channel_video_ids = get_vid_ids_by_channel_id(channel_id) 

    channel_subs = []
    for vid_id in channel_video_ids:
        split_subs = split_subtitles(vid_id[0])
        if split_subs is None:
            continue
        for sub in split_subs:
            start_time = sub[0]
            text = sub[1]
            embedding_subs = (channel_id, vid_id[0], start_time, text)
            channel_subs.append(embedding_subs)



    add_embeddings_to_chroma(channel_subs, openai_client)

    # mark the channel as enabled for semantic search 
    enable_ss(channel_id)
    console.print("[green]Embeddings generated[/green]")


# connfig
@cli.command(
    help = """
    Show config settings
    """
)
def config():
    config_path = get_config_path()
    db_path = get_db_path()
    chroma_path = get_or_make_chroma_path()

    console = Console()

    config_path = get_config_path()

    console.print(f"\nConfig directory: {config_path}\n")
    console.print(f"Database path: {db_path}\n")
    console.print(f"Chroma path: {chroma_path}\n")



from .download import get_channel_id_from_input
from .db_utils import * 
from .utils import *
from rich.console import Console


# full text search
def fts_search(text, scope, channel_id=None, video_id=None, limit=None):
    """
    Calls search functions and prints the results 
    """
    console = Console()

    if scope == "all":
        res = search_all(text, limit)
    
    if scope == "channel":
        channel_id = get_channel_id_from_input(channel_id)
        res = search_channel(channel_id, text, limit)
    
    if scope == "video":
        res = search_video(video_id, text, limit)

    if len(res) == 0:
        console.print("- Try shortening the search to specific words")
        console.print("- Try using the wildcard operator [bold]*[/bold] to search for partial words")
        console.print("- Try using the [bold]OR[/bold] operator to search for multiple words")
        if len(text.split(" ")) > 1:
            example_or = text.replace(" ", " OR ")
            console.print(f"    - EX: \"[bold]{example_or}[/bold]\"")
        else: 
            console.print(f"    - EX: \"[bold]foo OR [bold]bar[/bold]\"")
        exit()

    return res


# pretty print search results
def print_fts_res(res, query):

    console = Console()

    fts_res = []
    channel_names = []

    for quote in res:
        quote_match = {}
        video_id = quote["video_id"]
        time_stamp = quote["start_time"]
        time = time_to_secs(time_stamp)
        link = f"https://youtu.be/{video_id}?t={time}"

        quote_match["channel_name"] = get_channel_name_from_video_id(video_id)
        channel_names.append(quote_match["channel_name"])

        quote_match["video_title"] = get_title_from_db(video_id)
        quote_match["subs"] = bold_query_matches(quote["text"].strip(), query)
        quote_match["time_stamp"] = time_stamp
        quote_match["video_id"] = video_id
        quote_match["link"] = link 

        fts_res.append(quote_match)

    # sort by channel name
    fts_res = sorted(fts_res, key=lambda x: x["channel_name"])

    console.print("")
    for quote in fts_res: 

        console.print(f"[magenta][italic]\"[link={quote['link']}]{quote['subs']}[/link]\"[/italic][/magenta]")
        console.print(f"    Channel: {quote['channel_name']}",style="none")
        console.print(f"    Title: {quote['video_title']}")
        console.print(f"    Time Stamp: {quote['time_stamp']}")
        console.print(f"    Video ID: {quote['video_id']}")
        console.print(f"    Link: {quote['link']}")
        console.print("")
    
    num_matches = len(res)
    num_channels = len(set(channel_names))  
    num_videos = len(set([quote["video_id"] for quote in res]))

    summary_str = f"Found [bold]{num_matches}[/bold] matches in [bold]{num_videos}[/bold] videos from [bold]{num_channels}[/bold] channel"

    if num_channels > 1:
        summary_str += "s"

    console.print(summary_str) 


import chromadb

from .config import get_or_make_chroma_path

from openai import OpenAI
from rich.progress import track
from rich.console import Console


def add_embeddings_to_chroma(subs, openai_client):

    chroma_path = get_or_make_chroma_path()
    chroma_client = chroma_client = chromadb.PersistentClient(path=chroma_path) 
    collection = chroma_client.get_or_create_collection(name="subEmbeddings")

    for sub in track(subs, description="Getting embeddings"):
        channel_id = sub[0]
        video_id = sub[1]
        start_time = sub[2]
        text = sub[3]

        if text == '':
            continue

        embedding = get_embedding(text, "text-embedding-ada-002", openai_client)

        meta_data = {
            "channel_id": channel_id,
            "video_id": video_id,
            "start_time": start_time,
        }

        collection.add(
            documents=[text],
            embeddings=[embedding],
            metadatas=[meta_data],
            ids=[video_id + "_" + str(start_time)],
        )



def get_embedding(text, model="text-embedding-ada-002", client=None):
   if client is None:
       client = OpenAI()
   text = text.replace("\n", " ")
   return client.embeddings.create(input = [text], model=model).data[0].embedding




import csv, datetime, os

from rich.console import Console

from .db_utils import (
    search_channel, search_video, search_all, 
    get_channel_name_from_video_id, get_title_from_db
    )

from .utils import time_to_secs, show_message

def export_fts(text, scope, channel_id=None, video_id=None):
    """
    Calls search functions and exports the results to a csv file
    """

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    if scope == "all":
        file_name = f"all_{timestamp}.csv"
        res = search_all(text)
    if scope == "video":
        file_name = f"video_{video_id}_{timestamp}.csv"
        res = search_video(video_id, text)
    if scope == "channel":
        from .download import get_channel_id_from_input
        channel_id = get_channel_id_from_input(channel_id)
        file_name = f"channel_{channel_id}_{timestamp}.csv"
        res = search_channel(channel_id, text)


    if len(res) == 0:
        show_message("no_matches_found")
        return None

    with open(file_name, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(['Channel Name','Video Title', 'Quote', 'Time Stamp', 'Link'])
        
        for quote in res:
            video_id = quote["video_id"]
            channel_name = get_channel_name_from_video_id(video_id)
            video_title = get_title_from_db(video_id)
            time_stamp = quote["start_time"]
            subs = quote["text"]
            time = time_to_secs(time_stamp) 

            writer.writerow([channel_name,video_title, subs.strip(), time_stamp, f"https://youtu.be/{video_id}?t={time}"])
    
    console = Console()

    console.print(f"[bold]{len(res)}[/bold] matches found for text: \"[italic]{text}[/italic]\"")
    console.print(f"Exported to [green][bold]{file_name}[/bold][/green]")


def export_vector_search(res, search, scope):

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # run semantic search based on scope
    if scope == "all":
        file_name = f"all_{timestamp}.csv"
    if scope == "video":
        file_name = f"video_{timestamp}.csv"
    if scope == "channel":
        channel_id = res[0]["channel_id"]
        file_name = f"channel_{channel_id}_{timestamp}.csv"

    with open(file_name, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(['Channel Name','Video Title', 'Quote', 'Time Stamp', 'Link'])
        
        for quote in res:
            channel_name = quote["channel_name"] 
            video_title = quote["video_title"] 
            time_stamp = quote["start_time"]
            subs = quote["subs"]
            link = quote["link"]

            writer.writerow([channel_name,video_title, subs.strip(), time_stamp, link])
    
    console = Console()

    console.print(f"[bold]{len(res)}[/bold] matches found for text: \"[italic]{search}[/italic]\"")
    console.print(f"Exported to [green][bold]{file_name}[/bold][/green]")



def export_transcripts(channel_id):
    """
    Exports video transcripts from a channel to a text file 
    """

    console = Console()

    from .download import get_channel_id_from_input
    channel_id = get_channel_id_from_input(channel_id)

    from .db_utils import get_vid_ids_by_channel_id, get_transcript_by_video_id 
    videos = get_vid_ids_by_channel_id(channel_id)

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    file_name = f"channel_{channel_id}_{timestamp}.csv"

    for video in videos:

        video_id = video[0]
        transcript = get_transcript_by_video_id(video_id)
        str_transcript = ""
        for i in transcript:
            str_transcript += i[0] + "\n"
        with open(f"{video_id}.txt", "w") as f:
            f.write(str_transcript)


def export_channel_to_txt(channel_id):
    from .db_utils import  get_vid_ids_by_channel_id, get_subs_by_video_id
    console = Console()

    output_dir = f"{channel_id}_txt"

    if not os.path.exists(output_dir):
        os.mkdir(output_dir) 
    else:
        console.print(f"[red]Erorr:[/red] Directory [yellow]{output_dir}[/yellow] already exists")
        return None

    vid_ids = get_vid_ids_by_channel_id(channel_id)

    for vid_id in vid_ids:
        vid_id = vid_id[0]
        subs = get_subs_by_video_id(vid_id)
        str_subs = ""
        for sub in subs:
            str_subs += sub[2] + "\n"
        with open(f"{output_dir}/{vid_id}.txt", "w") as f:
            f.write(str_subs)

    return output_dir


def export_channel_to_vtt(channel_id):
    console = Console()
    from .db_utils import  get_vid_ids_by_channel_id, get_subs_by_video_id

    output_dir = f"{channel_id}_vtt"
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    else:
        console.print(f"[red]Erorr:[/red] Directory [yellow]{output_dir}[/yellow] already exists")
        return None



    vid_ids = get_vid_ids_by_channel_id(channel_id)

    for vid_id in vid_ids:
        vid_id = vid_id[0]
        subs = get_subs_by_video_id(vid_id)

        with open(f"{output_dir}/{vid_id}.vtt", "w") as f:
            f.write("WEBVTT\n\n")

        for sub in subs:
            start_time = sub[0]
            end_time = sub[1]
            text = sub[2]

            with open(f"{output_dir}/{vid_id}.vtt", "a") as f:
                f.write(f"{start_time} --> {end_time}\n{text}\n\n")

    return output_dir


"""
This is where I'm putting all the functions that don't belong anywhere else
"""
import re
import sqlite3

def show_message(code):
    error_dict = {
        "search_too_long": "Error: Search text must be less than 40 characters",
        "no_matches_found": "No matches found.\n- Try shortening the search text or use wildcards to match partial words.",
        "channel_not_found": "channel not found.\n- Try using channel id",
        "multiple_channels_found": "Multiple channels found.\n- Try using id",
        "channel_url_not_correct": "The given channel URL is not correct, expected pattern : https://www.youtube.com/@TimDillonShow/videos",
    }

    print(error_dict[code])


def time_to_secs(time_str):
    """
    converts timestamp to seconds youtube urls. Subtracts 3 seconds to give a buffer. 
    """
    time_rex = re.search("^(\d\d):(\d\d):(\d\d)",time_str)
    hours = int(time_rex.group(1)) * 3600 
    mins = int(time_rex.group(2)) * 60
    secs = int(time_rex.group(3)) 
    total_secs =  hours + mins + secs

    return total_secs - 3


def parse_vtt(file_path):
    """
    extracts start time and text from vtt file and return a list of dicts
    """
    result = []

    time_pattern = "^(.*) align:start position:0%"

    with open(file_path, "r") as f:
        lines = f.readlines()

    for count, line in enumerate(lines):
        time_match = re.match(time_pattern, line)

        if time_match:
            start = re.search("^(.*) -->",time_match.group(1))
            start_time = start.group(1)

            stop = re.search("--> (.*)",time_match.group(1))
            stop_time = stop.group(1)

            sub_titles = lines[count + 1]

            # prevent duplicate entries
            if result and result[-1]['text'] == sub_titles.strip('\n'):
                # replace the previous entry with the new one
                result[-1] = {
                    'start_time': start_time,
                    'stop_time': stop_time,
                    'text': sub_titles.strip('\n'),
                }
            else:   
                result.append({
                    'start_time': start_time,
                    'stop_time': stop_time,
                    'text': sub_titles.strip('\n'),
                })

    return result 


def get_api_key():
    import os
    api_key = os.environ.get("OPENAI_API_KEY")

    if api_key is None:
        return None
    return api_key


def get_time_delta(timestamp1, timestamp2):
    from datetime import datetime
    format_string = "%H:%M:%S.%f"
    dt1 = datetime.strptime(timestamp1, format_string)
    dt2 = datetime.strptime(timestamp2, format_string)
    diff = dt2 - dt1
    # convert to string "HH:MM:SS"
    diff = str(diff).split(".")[0]

    return diff 


# check if semantic search has been enabled for channel
def check_ss_enabled(channel_id=None):

    from yt_fts.config import get_db_path

    con = sqlite3.connect(get_db_path())
    cur = con.cursor()

    if channel_id is None:
        cur.execute(""" 
            SELECT channel_id FROM SemanticSearchEnabled 
            """)
    else:
        cur.execute(""" 
            SELECT channel_id FROM SemanticSearchEnabled 
            WHERE channel_id = ?
            """, [channel_id])

    res = cur.fetchone()
    if res is None:
        return False 
    else:
        return True 


# enable semantic search for channel
def enable_ss(channel_id):
    from yt_fts.config import get_db_path

    con = sqlite3.connect(get_db_path())
    cur = con.cursor()

    cur.execute(""" 
        INSERT INTO SemanticSearchEnabled (channel_id)
        VALUES (?)
        """, [channel_id])
    con.commit()
    con.close() 


def split_subtitles(video_id):

    from datetime import datetime
    from .db_utils import get_subs_by_video_id

    def time_to_seconds(time_str):
        """ Convert time string to total seconds """
        return datetime.strptime(time_str, '%H:%M:%S.%f').time().hour * 3600 + \
            datetime.strptime(time_str, '%H:%M:%S.%f').time().minute * 60 + \
            datetime.strptime(time_str, '%H:%M:%S.%f').time().second + \
            datetime.strptime(time_str, '%H:%M:%S.%f').time().microsecond / 1e6


    subs = get_subs_by_video_id(video_id)

    if len(subs) == 0:
        print("Video is too short to split")
        return None

    total_seconds = time_to_secs(subs[-1][1])

    if (total_seconds < 10):
        print("Video is too short to split")
    


    # Convert times to seconds and store texts
    converted_data = [(time_to_seconds(start), start, text) for start, end, text in subs]

    interval_texts = {}
    for start, start_time_str, text in converted_data:
        interval = int(start // 10) * 10  
        key = interval_texts.setdefault(interval, {'start_time': start_time_str, 'texts': []})
        key['texts'].append(text)


    result = [(data['start_time'], ' '.join(data['texts']).strip()) for data in interval_texts.values()]
    return result


def bold_query_matches(text, query):
    """
    Bold the query in the text, keeping the case the same
    """
    query_words = query.lower().split()
    result_words = []

    for word in text.split():
        if word.lower() in query_words:
            result_words.append(f"[bold]{word}[/bold]")
        else:
            result_words.append(word)

    return ' '.join(result_words)



import chromadb

from rich.console import Console
from sqlite_utils import Database

from .utils import time_to_secs, bold_query_matches
from .embeddings import get_embedding
from .config import get_or_make_chroma_path 
from .db_utils import get_channel_name_from_video_id, get_title_from_db
from .download import get_channel_id_from_input 

def search_chroma_db(
        text, 
        scope, 
        channel_id=None, 
        video_id=None, 
        limit=10,
        openai_client=None):

    chroma_path = get_or_make_chroma_path()
    chroma_client = chromadb.PersistentClient(path=chroma_path)
    collection = chroma_client.get_collection(name="subEmbeddings")

    search_embedding = get_embedding(text, "text-embedding-ada-002", openai_client)

    scope_options = {}

    if scope == "all":
        scope_options = {}
    if scope == "channel":
        scope_options = {"channel_id": get_channel_id_from_input(channel_id)}
    if scope == "video":
        scope_options = {"video_id": video_id}

    chroma_res = collection.query(
        query_embeddings=[search_embedding],
        n_results=limit,
        where=scope_options,
        )


    documents = chroma_res["documents"][0]
    metadata = chroma_res["metadatas"][0]
    distances = chroma_res["distances"][0]  

    res = []

    for i in range(len(documents)):
        text = documents[i]
        video_id = metadata[i]["video_id"]
        start_time = metadata[i]["start_time"]
        link = f"https://youtu.be/{video_id}?t={time_to_secs(start_time)}"
        channel_name = get_channel_name_from_video_id(video_id)
        channel_id = metadata[i]["channel_id"]
        title = get_title_from_db(video_id)


        match = {
            "distance": distances[i],
            "channel_name": channel_name,
            "channel_id": channel_id, 
            "video_title": title,
            "subs": text,
            "start_time": start_time,
            "video_id": video_id,
            "link": link,
        }
        res.append(match)

    return res


def print_vector_search_results(res, query):
    """
    {
    'channel_name': 'Peter Whidden', 
    'video_title': 'Training AI to Play Pokemon with Reinforcement Learning - YouTube', 
    'subs': 'choosing a Pokemon and winning its first', 
    'start_time': '00:22:14.909', 
    'video_id': 'DcYLT37ImBY', 
    'link': 'https://youtu.be/DcYLT37ImBY?t=1331'}
    """
    console = Console()

    
    channel_names = []

    for match in reversed(res):
        distance = match["distance"]
        link = match["link"]
        text = bold_query_matches(match["subs"], query)
        time_stamp = match["start_time"]    
        channel_id = match["channel_id"]
        video_id = match["video_id"]
        title = match["video_title"]
        channel_name = match["channel_name"]
        channel_names.append(channel_name)


        console.print(f"[magenta][italic]\"[link={link}]{text}[/link]\"[/italic][/magenta]\n")
        console.print(f"    Distance: {distance}",style="none")
        console.print(f"    Channel: {channel_name} - ({channel_id})",style="none")
        console.print(f"    Title: {title}")
        console.print(f"    Time Stamp: {time_stamp}")
        console.print(f"    Video ID: {video_id}")
        console.print(f"    Link: {link}")
        console.print("")

    num_matches = len(res)
    num_channels = len(set(channel_names))  
    num_videos = len(set([quote["video_id"] for quote in res]))

    summary_str = f"Found [bold]{num_matches}[/bold] matches in [bold]{num_videos}[/bold] videos from [bold]{num_channels}[/bold] channel"

    if num_channels > 1:
        summary_str += "s"

    console.print(summary_str) 



def delete_channel_from_chroma(channel_id):
    chroma_path = get_or_make_chroma_path()
    chroma_client = chromadb.PersistentClient(path=chroma_path)
    collection = chroma_client.get_collection(name="subEmbeddings")

    print(f"deleting channel {channel_id} from chroma")
    collection.delete(
        where={"channel_id": channel_id}
        )


import unittest
import os
import shutil

from pprint import pprint
from yt_fts.utils import parse_vtt

class TestParseVtt(unittest.TestCase):
    def test_parse_vtt(self):
        vtt_path = "tests/test_data/vtt/dqGyCTbzYmc.vtt"
        result = parse_vtt(vtt_path)

        pprint(result)


if __name__ == '__main__':
    unittest.main()


import unittest
import datetime
import os
from yt_fts.export import export_transcripts

class TestExportTranscripts(unittest.TestCase):
    def test_export_transcripts(self):
        # Setup
        channel_id = 2

        # Call the function with the test parameters
        export_transcripts(channel_id)

        # Construct the expected file name
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        expected_file_name = f"channel_{channel_id}_{timestamp}.csv"

        # Check that the file was created
        self.assertTrue(os.path.exists(expected_file_name))

        # Add assertions here to check the contents of the file
        # For example, you might open the file and check that it contains the expected transcripts

if __name__ == '__main__':
    unittest.main()


import subprocess, re, os, sqlite3, json

from concurrent.futures import ThreadPoolExecutor
from bs4 import BeautifulSoup
from urllib.parse import urlparse

from .config import get_db_path
from .db_utils import add_video
from .utils import parse_vtt
from urllib.parse import urlparse 

from rich.progress import track
from rich.console import Console

def handle_reject_consent_cookie(channel_url, s):
    """
    Auto rejects the consent cookie if request is redirected to the consent page
    """
    r = s.get(channel_url)
    if "https://consent.youtube.com" in r.url:
        m = re.search(r"<input type=\"hidden\" name=\"bl\" value=\"([^\"]*)\"", r.text)
        if m:
            data = {
                "gl":"DE",
                "pc":"yt",
                "continue":channel_url,
                "x":"6",
                "bl":m.group(1),
                "hl":"de",
                "set_eom":"true"
            }
            s.post("https://consent.youtube.com/save", data=data)


def get_channel_id(url, s):
    """
    Scrapes channel id from the channel page
    """
    res = s.get(url)
    if res.status_code == 200:
        html = res.text
        channel_id = re.search('channelId":"(.{24})"', html).group(1)
        return channel_id
    else:
        return None


def get_channel_name(channel_id, s):
    """
    Scrapes channel name from the channel page
    """
    console = Console()
    res = s.get(f"https://www.youtube.com/channel/{channel_id}/videos")

    if res.status_code == 200:

        html = res.text
        soup = BeautifulSoup(html, 'html.parser')
        script = soup.find('script', type='application/ld+json')

        # "Hot fix" for channels with special characters in the name
        try:
            with console.status("[bold green]Parsing JSON...") as status:
                data = json.loads(script.string)
        except:
            print("json parse failed retrying with escaped backslashes")
            script = script.string.replace('\\', '\\\\')
            data = json.loads(script)

        channel_name = data['itemListElement'][0]['item']['name']
        return channel_name 
    else:
        print("we couldn't get the channel name")
        return None



def get_videos_list(channel_url):
    """
    Scrapes list of all video urls from the channel
    """
    console = Console()

    with console.status("[bold green]Scraping video urls, this might take a little...") as status:
        cmd = [
            "yt-dlp",
            "--flat-playlist",
            "--print",
            "id",
            f"{channel_url}"
        ]
        res = subprocess.run(cmd, capture_output=True, check=True)
        list_of_videos_urls = res.stdout.decode().splitlines()

        streams_url = channel_url.replace("/videos", "/streams") 
        cmd = [
            "yt-dlp",
            "--flat-playlist",
            "--print",
            "id",
            streams_url
        ]
        try:
            res = subprocess.run(cmd, capture_output=True, check=True)
            live_stream_urls = res.stdout.decode().splitlines()
            if len(live_stream_urls) > 0:
                list_of_videos_urls.extend(live_stream_urls)
        except subprocess.CalledProcessError:
            console.print("[bold red]No streams tab found or error fetching streams.")


    return list_of_videos_urls 


def download_vtts(number_of_jobs, list_of_videos_urls, language ,tmp_dir):
    """
    Multi-threaded download of vtt files
    """

    # showing progress on a multi-threaded task might be more trouble than it's worth
    # console = Console()

    executor = ThreadPoolExecutor(number_of_jobs)
    futures = []

    for video_id in list_of_videos_urls:
        video_url = f'https://www.youtube.com/watch?v={video_id}'
        future = executor.submit(get_vtt, tmp_dir, video_url, language)
        futures.append(future)
    
    for i in range(len(list_of_videos_urls)):
        futures[i].result()


def get_vtt(tmp_dir, video_url, language):
    subprocess.run([
        "yt-dlp",
        "-o", f"{tmp_dir}/%(id)s",
        "--write-info-json",
        "--write-auto-sub",
        "--convert-subs", "vtt",
        "--skip-download",
        "--sub-langs", f"{language},-live_chat",
        video_url
    ])


def vtt_to_db(channel_id, dir_path, s):
    """
    Iterates through all vtt files in the temp_dir, passes them to 
    the vtt parsing function, then inserts the data into the database.
    """
    items = os.listdir(dir_path)
    file_paths = []

    for item in items:
        # ignore other files e.g. info.json files
        if not item.endswith('.vtt'):
            continue

        item_path = os.path.join(dir_path, item)
        if os.path.isfile(item_path):
            file_paths.append(item_path)    

    con = sqlite3.connect(get_db_path())  
    cur = con.cursor()

    # bar = Bar('Adding to database', max=len(file_paths))

    for vtt in track(file_paths, description="Adding subtitles to database..."):
        base_name = os.path.basename(vtt)
        vid_id = base_name.split('.')[0]
        vid_url = f"https://youtu.be/{vid_id}"
        vid_title = get_vid_title(os.path.join(os.path.dirname(vtt), f'{vid_id}.info.json'))
        add_video(channel_id, vid_id, vid_title, vid_url)

        vtt_json = parse_vtt(vtt)

        for sub in vtt_json:
            start_time = sub['start_time']
            stop_time = sub['stop_time']
            text = sub['text']
            cur.execute(f"INSERT INTO Subtitles (video_id, start_time, stop_time, text) VALUES (?, ?, ?, ?)", 
                        (vid_id, start_time, stop_time, text))

        con.commit()

    con.close()


def get_vid_title(info_json_path):
    """
    Retrieves video title from the info json file.
    """
    with open(info_json_path) as f:
        return json.load(f)['title']


def validate_channel_url(channel_url):
    """
    valid patterns
    https://www.youtube.com/channel/channelID
    https://www.youtube.com/@channelname

    https://www.youtube.com/@channelname/*
    https://www.youtube.com/channel/channelID/*
    """
    from rich.console import Console
    console = Console()

    channel_url = channel_url.strip('/')
    parsed = urlparse(channel_url)
    domain = parsed.netloc
    path = parsed.path.split('/')

    if not domain.endswith('youtube.com'):
        console.print("")
        console.print(f"[bold red]Error:[/bold red] Invalid channel domain: [blue]{domain}[/blue]")
        console.print("")
        exit()
    
    if len(path) < 2:
        console.print("")
        console.print(f"[bold red]Error:[/bold red] Invalid channel url: [blue]{channel_url}[/blue]")
        console.print("")
        exit()

    if path[1].startswith("@"):
        return f"https://www.youtube.com/{path[1]}/videos"

    if path[1] == "channel":
        return f"https://www.youtube.com/channel/{path[2]}/videos"

    console.print("")
    console.print(f"[bold red]Error:[/bold red] Unknown url format: [blue]{channel_url}[/blue]")
    console.print("")
    console.print("Please use one of the following formats:")
    console.print("")
    console.print("    - https://www.youtube.com/[yellow]@channelname[/yellow]")
    console.print("    - https://www.youtube.com/channel/[yellow]channelId[/yellow]")
    console.print("")
    exit()


def download_channel(channel_id, channel_name, language, number_of_jobs, s):
    """
    Downloads all the videos from a channel to a tmp directory
    """

    import tempfile
    from yt_fts.db_utils import add_channel_info

    console = Console() 

    with tempfile.TemporaryDirectory() as tmp_dir:
        channel_url = f"https://www.youtube.com/channel/{channel_id}/videos"
        list_of_videos_urls = get_videos_list(channel_url)

        console.print(f"[green][bold]Downloading [red]{len(list_of_videos_urls)}[/red] vtt files[/bold][/green]\n")
        console.print("[green]I would normally show a progress bar here, but multithreading and progress bars don't play nice.[/green]\n")

        download_vtts(number_of_jobs, list_of_videos_urls, language, tmp_dir)
        add_channel_info(channel_id, channel_name, channel_url)
        vtt_to_db(channel_id, tmp_dir, s)
    return True


def get_channel_id_from_input(channel_input):
    """
    Checks if the input is a rowid or a channel name and returns channel id
    """

    from yt_fts.db_utils import (
        get_channel_id_from_rowid, 
        get_channel_id_from_name
    )

    from yt_fts.utils import show_message

    name_res = get_channel_id_from_name(channel_input) 
    id_res = get_channel_id_from_rowid(channel_input) 


    
    if id_res != None:
        return id_res
    elif name_res != None: 
        return name_res
    else:
        show_message("channel_not_found")
        exit()
    